import hashlib
import json
import logging
import os
from typing import Optional, Protocol

import websockets


CHUNK_SIZE = 5 * 1024 * 1024  # 5 MiB


class StreamReporter(Protocol):
    """Reporter interface to decouple streaming output from stdout."""

    def info(self, message: str) -> None: ...

    def warning(self, message: str) -> None: ...

    def error(self, message: str) -> None: ...

    def progress(
        self,
        current: int,
        total: int,
        prefix: str = "",
        suffix: str = "",
        length: int = 40,
    ) -> None: ...


class LoggingReporter:
    """Default reporter using the Python logging module."""

    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.logger = logger or logging.getLogger(__name__)

    def info(self, message: str) -> None:
        self.logger.info(message)

    def warning(self, message: str) -> None:
        self.logger.warning(message)

    def error(self, message: str) -> None:
        self.logger.error(message)

    def progress(
        self,
        current: int,
        total: int,
        prefix: str = "",
        suffix: str = "",
        length: int = 40,
    ) -> None:
        # Logging progress as debug avoids log spam when callers do not care.
        percent = 100 * (current / float(total))
        self.logger.debug("%s %.1f%% %s", prefix, percent, suffix)


class ConsoleReporter:
    """Reporter that prints to stdout with a simple progress bar."""

    def info(self, message: str) -> None:
        print(message)

    def warning(self, message: str) -> None:
        print(message)

    def error(self, message: str) -> None:
        print(message)

    def progress(
        self,
        current: int,
        total: int,
        prefix: str = "",
        suffix: str = "",
        length: int = 40,
    ) -> None:
        printProgressBar(current, total, prefix=prefix, suffix=suffix, length=length)


# Print iterations progress
def printProgressBar(
    iteration,
    total,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="█",
    printEnd="\r",
):
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)
    # Print New Line on Complete
    if iteration == total:
        print()


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


async def stream_send(websocket, target, reporter: Optional[StreamReporter] = None):
    reporter = reporter or LoggingReporter()
    try:
        with open(target, "rb") as f:
            file_size = os.stat(target).st_size
            num_chunks = (file_size + CHUNK_SIZE - 1) // CHUNK_SIZE
            hash = hashlib.new("sha256")
            try:
                await websocket.send(
                    json.dumps(
                        {
                            "type": "init",
                            "num_chunks": num_chunks,
                            "file_size": file_size,
                        }
                    ).encode(encoding="utf-8")
                )
                reporter.info(f"Transfering {sizeof_fmt(file_size)}...")
                chunk_count = 0
                while chunk := f.read(CHUNK_SIZE):
                    await websocket.send(chunk, text=False)
                    hash.update(chunk)
                    chunk_count += 1
                    reporter.progress(chunk_count, num_chunks, length=40)
                await websocket.send(
                    json.dumps({"type": "eof", "sha256_hash": hash.hexdigest()}).encode(
                        encoding="utf-8"
                    )
                )
                reporter.info(f"File {target} sent successfully.")
                return True
            except websockets.exceptions.ConnectionClosedError as e:
                reporter.error(f"Remote connection closed with error: {e}")
                return False
    except FileNotFoundError:
        reporter.error(f"File {target} not found. Aborting transfer.")
        await websocket.close(
            code=3003,
            reason=json.dumps({"type": "error", "error": "FileNotFoundError"}),
        )
        return False
    except Exception as e:
        reporter.error(f"An error occurred: {e}")
        await websocket.close(
            code=1011,
            reason=json.dumps({"type": "error", "error": type(e).__name__}),
        )
        return False


async def stream_receive(
    websocket, target, size_limit=None, reporter: Optional[StreamReporter] = None
):
    reporter = reporter or LoggingReporter()
    try:
        init = None
        transfer_size = 0
        with open(target, "xb") as f:
            hash = hashlib.new("sha256")
            chunk_count = 0
            async for message in websocket:
                if init is None:
                    init = json.loads(message.decode("utf-8"))
                    reporter.info(f"Transfering {sizeof_fmt(init['file_size'])}...")
                    continue
                if isinstance(message, str) and message.startswith('{"type":"error"'):
                    error = json.loads(message.decode("utf-8"))
                    reporter.error(f"A remote error occurred: {error['error']}")
                    break
                if isinstance(message, str) and message.startswith('{"type":"eof"'):
                    if hash.hexdigest() != json.loads(message)["sha256_hash"]:
                        reporter.error("Hash mismatch! File transfer corrupted.")
                        os.remove(target)
                    else:
                        reporter.info("File received successfully.")
                    break
                transfer_size += CHUNK_SIZE
                if size_limit is not None and transfer_size > size_limit:
                    reporter.error(
                        f"Inbound transfer limit exceeded, max allowed transfer size: {size_limit} bytes Aborting transfer."
                    )
                    await websocket.close(
                        code=1009,
                        reason=f"Inbound transfer limit exceeded, max allowed transfer size: {size_limit} bytes.",
                    )
                    os.remove(target)
                    return False
                f.write(message)
                hash.update(message)
                chunk_count += 1
                reporter.progress(chunk_count, init["num_chunks"], length=40)
            reporter.info(f"File {target} received successfully.")
            return True
    except FileExistsError:
        reporter.error(f"File {target} already exists. Transfer aborted.")
        await websocket.close(
            code=3003,
            reason=json.dumps({"type": "error", "error": "FileNotFoundError"}),
        )
        return False
    except Exception as e:
        reporter.error(f"An error occurred: {e}.")
        await websocket.close(
            code=1011,
            reason=json.dumps({"type": "error", "error": type(e).__name__}),
        )
        return False



import asyncio
import base64
import click
import json
import websockets
from dataclasses import dataclass
from typing import Optional
from streamer.streamer_core import (
    ConsoleReporter,
    LoggingReporter,
    StreamReporter,
    stream_send,
    stream_receive,
)

CHUNK_SIZE = 5 * 1024 * 1024  # 5 MiB


@dataclass
class ClientConfig:
    target: str
    port_range: list[int]
    ip_list: list[str]
    secret: str


async def client_receive(
    config: ClientConfig, reporter: Optional[StreamReporter] = None
):
    reporter = reporter or LoggingReporter()
    try:
        for ip in config.ip_list:
            for port in range(config.port_range[0], config.port_range[1] + 1):
                uri = f"ws://{ip}:{port}"
                try:
                    async with websockets.connect(
                        uri,
                        max_size=int(
                            CHUNK_SIZE * 1.25
                        ),  # Allow some overhead for encoding and headers
                        ping_interval=60,
                        ping_timeout=None,
                        additional_headers={"Authorization": f"Bearer {config.secret}"},
                    ) as websocket:
                        await stream_receive(
                            websocket, config.target, reporter=reporter
                        )
                        return
                except (
                    OSError,
                    websockets.exceptions.InvalidStatus,
                    websockets.exceptions.InvalidMessage,
                ):
                    continue
        reporter.error("Unable to establish connection to any provided IPs/ports.")
    except Exception as e:
        reporter.error(f"An error occurred: {e}")
        return


async def client_send(config: ClientConfig, reporter: Optional[StreamReporter] = None):
    reporter = reporter or LoggingReporter()
    try:
        for ip in config.ip_list:
            for port in range(config.port_range[0], config.port_range[1] + 1):
                uri = f"ws://{ip}:{port}"
                try:
                    async with websockets.connect(
                        uri,
                        max_size=int(
                            CHUNK_SIZE * 1.25
                        ),  # Allow some overhead for encoding and headers
                        ping_interval=60,
                        ping_timeout=None,
                        additional_headers={"Authorization": f"Bearer {config.secret}"},
                    ) as websocket:
                        await stream_send(websocket, config.target, reporter=reporter)
                        return
                except (
                    OSError,
                    websockets.exceptions.InvalidStatus,
                    websockets.exceptions.InvalidMessage,
                ):
                    continue
        reporter.error("Unable to establish connection to any provided IPs/ports.")
    except Exception as e:
        reporter.error(f"An error occurred: {e}")
        return


def set_coordinates(coordinates) -> ClientConfig:
    try:
        json_str = base64.urlsafe_b64decode(coordinates).decode("utf-8")
        data = json.loads(json_str)

        return ClientConfig(
            target="",
            secret=data["secret"],
            port_range=data["ports"],
            ip_list=data["ips"],
        )
    except (json.JSONDecodeError, KeyError, base64.binascii.Error) as e:
        raise click.ClickException("Invalid coordinates format") from e


@click.command()
@click.option("--path", help="The source path of the file to be sent.", required=True)
@click.option(
    "--coordinates",
    help="Secret coordinates used to establish a connection",
    required=True,
)
def send(path, coordinates):
    config = set_coordinates(coordinates)
    config.target = path
    asyncio.run(client_send(config, reporter=ConsoleReporter()))


@click.command()
@click.option(
    "--coordinates",
    help="Secret coordinates used to establish a connection",
    required=True,
)
@click.option("--path", help="The target path of the incoming file.", required=True)
def receive(path, coordinates):
    config = set_coordinates(coordinates)
    config.target = path
    asyncio.run(client_receive(config, reporter=ConsoleReporter()))



import asyncio
import base64
import click
import http
import json
import signal
import websockets
from enum import Enum
from dataclasses import dataclass, replace
from streamer.streamer_core import ConsoleReporter, stream_receive, stream_send
from websockets.asyncio.server import serve

CHUNK_SIZE = 5 * 1024 * 1024  # 5 MiB


class Operation(Enum):
    send = "send"
    receive = "receive"


@dataclass
class StreamConfig:
    operation: str
    target: str
    secret: str
    port_range: tuple[int, int]
    ips: list[str]
    host: str
    wait_timeout: int
    inbound_transfer_limit: int


async def stream(config: StreamConfig):
    timeout_handle: asyncio.Handle = None
    reporter = ConsoleReporter()

    def process_request(connection, request):
        nonlocal timeout_handle
        if "Authorization" not in request.headers:
            return connection.respond(
                http.HTTPStatus.UNAUTHORIZED, "Missing Authorization header\n"
            )

        authorization = request.headers["Authorization"]
        if authorization is None:
            return connection.respond(http.HTTPStatus.UNAUTHORIZED, "Missing token\n")

        token = authorization.split("Bearer ")[-1]
        if token is None or token != config.secret:
            return connection.respond(http.HTTPStatus.FORBIDDEN, "Invalid secret\n")

        if timeout_handle:
            timeout_handle.cancel()

    async def server_receive(websocket: websockets.asyncio.server.ServerConnection):
        reporter.info("Client connected.")
        try:
            await stream_receive(
                websocket,
                config.target,
                config.inbound_transfer_limit,
                reporter=reporter,
            )
        except Exception as e:
            reporter.error(f"An error occurred: {e}")
        websocket.server.close()

    async def server_send(websocket):
        reporter.info("Client connected.")
        try:
            await stream_send(websocket, config.target, reporter=reporter)
        except Exception as e:
            reporter.error(f"An error occurred: {e}")
        websocket.server.close()

    start_port, end_port = config.port_range
    for port in range(start_port, end_port + 1):
        try:
            async with serve(
                (
                    server_receive
                    if config.operation == Operation.receive
                    else server_send
                ),
                config.host,
                port,
                max_size=int(
                    CHUNK_SIZE * 1.25
                ),  # Allow some overhead for encoding and headers
                ping_interval=60,
                ping_timeout=None,
                process_request=process_request,
            ) as server:
                reporter.info(f"Server is listening on ws://{config.host}:{port}")
                coordinates = {
                    "ports": [start_port, end_port],
                    "ips": config.ips,
                    "secret": config.secret,
                }
                encoded = base64.urlsafe_b64encode(
                    json.dumps(coordinates).encode("utf-8")
                ).decode("utf-8")

                reporter.info(f"Use these coordinates to connect: {encoded}")

                loop = asyncio.get_running_loop()
                loop.add_signal_handler(signal.SIGTERM, server.close)
                timeout_handle = loop.call_later(config.wait_timeout, server.close)
                await server.wait_closed()
            break
        except OSError:
            reporter.error(f"Server unable to bing on port: {port}")
            continue


@click.group()
@click.option(
    "--secret",
    "_secret",
    help="A shared secret required to initiate the transfer",
    required=True,
)
@click.option(
    "--public-ips",
    "_ips",
    help="A list of public IPs where the streamer server might run.",
    default=["localhost"],
    multiple=True,
)
@click.option(
    "--host",
    "_host",
    help="The interface to use for listening incoming connections",
    default="localhost",
)
@click.option(
    "--port-range",
    "_port_range",
    type=(int, int),
    help="A range of ports to pick from to listen for incoming connections e.g. --port-range 5665 5670",
    default=(5665, 5670),
)
@click.option(
    "--wait-timeout",
    "_wait_timeout",
    help="How long to wait for a connection before exiting (in seconds)",
    default=60 * 60 * 24,  # 24h
)
@click.option(
    "--inbound-transfer-limit",
    "_inbound_transfer_limit",
    help="Limit how much data can be received (in bytes)",
    default=5 * 1024 * 1024 * 1024,  # 5GB
)
@click.pass_context
def server(
    ctx, _secret, _ips, _host, _port_range, _wait_timeout, _inbound_transfer_limit
):
    ctx.ensure_object(dict)
    ctx.obj = StreamConfig(
        operation=None,
        target=None,
        secret=_secret,
        port_range=_port_range,
        ips=_ips,
        host=_host,
        wait_timeout=_wait_timeout,
        inbound_transfer_limit=_inbound_transfer_limit,
    )


@server.command()
@click.option("--path", help="The target path of the file to be sent.", required=True)
@click.pass_context
def send(ctx, path):
    config = replace(ctx.obj, operation=Operation.send, target=path)
    asyncio.run(stream(config))


@server.command()
@click.option(
    "--path", help="The target path of the file to be received.", required=True
)
@click.pass_context
def receive(ctx, path):
    config = replace(ctx.obj, operation=Operation.receive, target=path)
    asyncio.run(stream(config))
